# -*- coding: utf-8 -*-

"""Bio2BEL custom errors."""


class Bio2BELMissingNameError(TypeError):
    """Raised when an abstract manager is subclassed and instantiated without overriding the module name."""


class Bio2BELModuleCaseError(TypeError):
    """Raised when the module name in a subclassed and instantiated manager is not all lowercase."""


class Bio2BELMissingModelsError(TypeError):
    """Raises when trying to build a flask admin app with no models defined."""


class Bio2BELTestMissingManagerError(TypeError):
    """Raised when "Manager" was not set as a class-level variable.."""


class Bio2BELManagerTypeError(TypeError):
    """Raised when the class-level variable "Manager" is not a subclass of :class:`bio2bel.AbstractManager`."""



# -*- coding: utf-8 -*-

"""This module has tools for making packages that can reproducibly download and parse data."""

import logging
import os
from urllib.request import urlretrieve

import pandas as pd

log = logging.getLogger(__name__)

__all__ = [
    'make_downloader',
    'make_df_getter',
]


def make_downloader(url, path):
    """Make a function that downloads the data for you, or uses a cached version at the given path.

    :param str url: The URL of some data
    :param str path: The path of the cached data, or where data is cached if it does not already exist
    :return: A function that downloads the data and returns the path of the data
    :rtype: (bool -> str)
    """
    def download_data(force_download=False):
        """Download the data.

        :param bool force_download: If true, overwrites a previously cached file
        :rtype: str
        """
        if os.path.exists(path) and not force_download:
            log.info('using cached data at %s', path)
        else:
            log.info('downloading %s to %s', url, path)
            urlretrieve(url, path)

        return path

    return download_data


def make_df_getter(data_url, data_path, **kwargs):
    """Build a function that handles downloading tabular data and parsing it into a pandas DataFrame.

    :param str data_url: The URL of the data
    :param str data_path: The path where the data should get stored
    :param kwargs: Any other arguments to pass to :func:`pandas.read_csv`
    :rtype: (Optional[str], bool, bool) -> pandas.DataFrame
    """
    download_function = make_downloader(data_url, data_path)

    def get_df(url=None, cache=True, force_download=False):
        """Get the data as a pandas DataFrame.

        :param Optional[str] url: The URL (or file path) to download.
        :param bool cache: If true, the data is downloaded to the file system, else it is loaded from the internet
        :param bool force_download: If true, overwrites a previously cached file
        :rtype: pandas.DataFrame
        """
        if url is None and cache:
            url = download_function(force_download=force_download)

        return pd.read_csv(
            url or data_url,
            **kwargs
        )

    return get_df



# -*- coding: utf-8 -*-

"""Utilities for Bio2BEL."""

from configparser import ConfigParser
import logging
import os

from .constants import BIO2BEL_DIR, DEFAULT_CACHE_CONNECTION, DEFAULT_CONFIG_PATH, VERSION

log = logging.getLogger(__name__)

__all__ = [
    'get_data_dir',
    'get_connection',
    'get_version',
]


def get_data_dir(module_name):
    """Ensure the appropriate Bio2BEL data directory exists for the given module, then returns the file path.

    :param str module_name: The name of the module. Ex: 'chembl'
    :return: The module's data directory
    :rtype: str
    """
    module_name = module_name.lower()
    data_dir = os.path.join(BIO2BEL_DIR, module_name)
    os.makedirs(data_dir, exist_ok=True)
    return data_dir


def get_connection(module_name, connection=None):
    """Return the SQLAlchemy connection string if it is set.

    Order of operations:

    1. Return the connection if given as a parameter
    2. Check the environment for BIO2BEL_{module_name}_CONNECTION
    3. Look in the bio2bel config file for module-specific connection. Create if doesn't exist. Check the
       module-specific section for ``connection``
    4. Look in the bio2bel module folder for a config file. Don't create if doesn't exist. Check the default section
       for ``connection``
    5. Check the environment for BIO2BEL_CONNECTION
    6. Check the bio2bel config file for default
    7. Fall back to standard default cache connection

    :param str module_name: The name of the module to get the configuration for
    :param Optional[str] connection: get the SQLAlchemy connection string
    :return: The SQLAlchemy connection string based on the configuration
    :rtype: str
    """
    module_name = module_name.lower()

    # 1. Use given connection
    if connection is not None:
        return connection

    # 2. Check the environment for the module
    bio2bel_module_env = 'BIO2BEL_{}_CONNECTION'.format(module_name.upper())
    bio2bel_module_env_value = os.environ.get(bio2bel_module_env)
    if bio2bel_module_env_value is not None:
        log.debug('loaded connection from environment (%s): %s', bio2bel_module_env, bio2bel_module_env_value)
        return bio2bel_module_env_value

    # 4. Check the global Bio2BEL configuration for module-specific connection information
    global_config = ConfigParser()
    local_config = ConfigParser()

    if os.path.exists(DEFAULT_CONFIG_PATH):
        global_config.read(DEFAULT_CONFIG_PATH)
        if global_config.has_option(module_name, 'connection'):
            global_module_connection = global_config.get(module_name, 'connection')
            log.debug('loading connection string from global configuration (%s): %s', DEFAULT_CONFIG_PATH,
                      global_module_connection)
            return global_module_connection

    # 5. Check if there is module-specific configuration
    module_config_path = os.path.join(BIO2BEL_DIR, module_name, 'config.ini')
    if os.path.exists(module_config_path):
        local_config.read(module_config_path)
        if local_config.has_option(local_config.default_section, 'connection'):
            local_module_connection = local_config.get(local_config.default_section, 'connection')
            log.debug('loading connection string from local configuration (%s)', module_config_path,
                      local_module_connection)
            return local_module_connection

    # 6. Check if there is a global connection
    global_environ_connection = os.environ.get('BIO2BEL_CONNECTION')
    if global_environ_connection is not None:
        log.debug('loading global bio2bel connection from environ: %s', global_environ_connection)
        return global_environ_connection

    # 7. Use the global configuration file's global default cache connection string
    if not os.path.exists(DEFAULT_CONFIG_PATH):
        log.debug('creating config file: %s', DEFAULT_CONFIG_PATH)
        config_writer = ConfigParser()
        with open(DEFAULT_CONFIG_PATH, 'w') as file:
            config_writer.set(config_writer.default_section, 'connection', DEFAULT_CACHE_CONNECTION)
            config_writer.write(file)

    log.debug('fetching global bio2bel config from %s', DEFAULT_CONFIG_PATH)
    config = ConfigParser()
    config.read(DEFAULT_CONFIG_PATH)

    if not config.has_option(config.default_section, 'connection'):
        log.debug('creating default connection string %s', DEFAULT_CACHE_CONNECTION)
        return DEFAULT_CACHE_CONNECTION

    default_connection = config.get(config.default_section, 'connection')
    log.debug('load default connection string from %s', default_connection)

    return default_connection


def get_version():
    """Get the software version of Bio2BEL."""
    return VERSION



# -*- coding: utf-8 -*-

"""Provides abstractions over the management of SQLAlchemy connections and sessions."""

import os

import click

from .cli_manager import CliMixin
from .connection_manager import ConnectionManager
from ..exc import Bio2BELMissingModelsError

__all__ = [
    'FlaskMixin',
]


class FlaskMixin(ConnectionManager, CliMixin):
    """A mixin for building a Flask-Admin interface.

    This class can be used as a mixin, meaning that a class inheriting from AbstractManager can also multiple-inherit
    from this class. It contains functions to build a :mod:`flask` application for easy viewing of the contents of the
    database.

    First, you'll have to make sure that the :mod:`flask` and :mod:`flask-admin` are installed. They can be installed
    with Bio2BEL using the package extra called "web" like:

    .. code-block:: bash

        $ pip install bio2bel[web]

    Or, installed directly with `pip`:

    .. code-block:: bash

        $ pip install flask flask-admin

    Besides this, all that's necessary to use this mixin is to define the class variable ``flask_admin_models`` as a
    list of SQLAlchemy models you'd like to see.

    .. code-block:: python

        >>> from sqlalchemy.ext.declarative import DeclarativeMeta
        >>>
        >>> from bio2bel import AbstractManager
        >>> from bio2bel.manager.flask_manager.FlaskMixin
        >>>
        >>> from .constants import MODULE_NAME
        >>> from .models import Base, Evidence, Interaction, Mirna, Species, Target
        >>>
        >>> class Manager(AbstractManager):
        ...    module_name = MODULE_NAME
        ...    flask_admin_models = [Evidence, Interaction, Mirna, Species, Target]
        ...
        ...    @property
        ...    def _base(self) -> DeclarativeMeta:
        ...        return Base
        ...
        ...    def populate(self) -> None:
        ...        ...
    """

    #: Represents a list of SQLAlchemy classes to make a Flask-Admin interface.
    flask_admin_models = ...

    def __init__(self, *args, **kwargs):  # noqa: D107
        if self.flask_admin_models is ...:
            raise Bio2BELMissingModelsError(
                'FlaskMixin necessitates definition of class variable "flask_admin_models".')

        super().__init__(*args, **kwargs)

    def _add_admin(self, app, **kwargs):
        """Add a Flask Admin interface to an application.

        :param flask.Flask app: A Flask application
        :param kwargs: Keyword arguments are passed through to :class:`flask_admin.Admin`
        :rtype: flask_admin.Admin
        """
        from flask_admin import Admin
        from flask_admin.contrib.sqla import ModelView

        admin = Admin(app, **kwargs)

        for flask_admin_model in self.flask_admin_models:
            if isinstance(flask_admin_model, tuple):  # assume its a 2 tuple
                if len(flask_admin_model) != 2:
                    raise TypeError

                model, view = flask_admin_model
                admin.add_view(view(model, self.session))

            else:
                admin.add_view(ModelView(flask_admin_model, self.session))

        return admin

    def get_flask_admin_app(self, url=None, secret_key=None):
        """Create a Flask application.

        :param Optional[str] url: Optional mount point of the admin application. Defaults to ``'/'``.
        :rtype: flask.Flask
        """
        from flask import Flask

        app = Flask(__name__)

        if secret_key:
            app.secret_key = secret_key

        self._add_admin(app, url=(url or '/'))
        return app

    @staticmethod
    def _cli_add_flask(main):
        """Add the web command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_flask(main)

    @classmethod
    def get_cli(cls):
        """Add  a :mod:`click` main function to use as a command line interface.

        :rtype: click.Group
        """
        main = super().get_cli()

        cls._cli_add_flask(main)

        return main


def add_cli_flask(main):
    """Add a ``web`` comand main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-v', '--debug', is_flag=True)
    @click.option('-p', '--port')
    @click.option('-h', '--host')
    @click.option('-k', '--secret-key', default=os.urandom(8))
    @click.pass_obj
    def web(manager, debug, port, host, secret_key):
        """Run the web app."""
        app = manager.get_flask_admin_app(url='/', secret_key=secret_key)
        app.run(debug=debug, host=host, port=port)

    return main



# -*- coding: utf-8 -*-

"""Provides abstractions over the management of SQLAlchemy connections and sessions."""

import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from ..exc import Bio2BELMissingNameError, Bio2BELModuleCaseError
from ..models import Action, create_all
from ..utils import get_connection

log = logging.getLogger(__name__)

__all__ = [
    'ConnectionManager',
]


class ConnectionManager(object):
    """Represents the connection-building aspect of the abstract manager.

    Minimally requires the definition of the class-level variable, ``module_name``.

    Example for InterPro:

    >>> from bio2bel.manager import ConnectionManager
    >>> class Manager(ConnectionManager):
    >>>     module_name = 'interpro'

    In general, this class won't be used directly except in the situation where the connection should be loaded
    in a different way and it can be used as a mixin.
    """

    #: This represents the module name. Needs to be lower case
    module_name = ...

    def __init__(self, connection=None, engine=None, session=None, **kwargs):
        """Build an abstract manager from either a connection or an engine/session.

        The remaining keyword arguments are passed to :func:`build_engine_session`.

        :param Optional[str] connection:
        :param engine:
        :param session:
        """
        self._assert_module_name()

        if connection and (engine or session):
            raise ValueError('can not specify connection with engine/session')

        if engine is None and session is None:
            if connection is None:
                connection = self._get_connection()

            engine, session = build_engine_session(connection=connection, **kwargs)

        self.engine = engine
        self.session = session

    @property
    def connection(self):
        """Return this manager's connection string."""
        return str(self.engine.url)

    @classmethod
    def _assert_module_name(cls):
        if cls.module_name is ...:
            raise Bio2BELMissingNameError('module_name class variable not set on {}'.format(cls.__name__))

        if not isinstance(cls.module_name, str):
            raise TypeError('module_name class variable not set as str: {}'.format(cls.__name__))

        if cls.module_name != cls.module_name.lower():
            raise Bio2BELModuleCaseError('module_name class variable should be lowercase')

    @classmethod
    def _get_connection(cls, connection=None):
        """Get a default connection string.

        Wraps :func:`bio2bel.utils.get_connection` and passing this class's :data:`module_name` to it.

        :param Optional[str] connection: A custom connection to pass through
        :rtype: str
        """
        return get_connection(cls.module_name, connection=connection)

    def _store_populate(self):
        create_all(self.engine)
        Action.store_populate(self.module_name, session=self.session)

    def _store_drop(self):
        create_all(self.engine)
        Action.store_drop(self.module_name, session=self.session)

    def __repr__(self):  # noqa: D105
        return '<{module_name}Manager url={url}>'.format(
            module_name=self.module_name.capitalize(),
            url=self.engine.url
        )


def build_engine_session(connection, echo=False, autoflush=None, autocommit=None, expire_on_commit=None,
                         scopefunc=None):
    """Build an engine and a session.

    :param str connection: An RFC-1738 database connection string
    :param bool echo: Turn on echoing SQL
    :param Optional[bool] autoflush: Defaults to True if not specified in kwargs or configuration.
    :param Optional[bool] autocommit: Defaults to False if not specified in kwargs or configuration.
    :param Optional[bool] expire_on_commit: Defaults to False if not specified in kwargs or configuration.
    :param scopefunc: Scoped function to pass to :func:`sqlalchemy.orm.scoped_session`
    :rtype: tuple[Engine,Session]

    From the Flask-SQLAlchemy documentation:

    An extra key ``'scopefunc'`` can be set on the ``options`` dict to
    specify a custom scope function.  If it's not provided, Flask's app
    context stack identity is used. This will ensure that sessions are
    created and removed with the request/response cycle, and should be fine
    in most cases.
    """
    if connection is None:
        raise ValueError('can not build engine when connection is None')

    engine = create_engine(connection, echo=echo)

    autoflush = autoflush if autoflush is not None else False
    autocommit = autocommit if autocommit is not None else False
    expire_on_commit = expire_on_commit if expire_on_commit is not None else True

    log.debug('auto flush: %s, auto commit: %s, expire on commmit: %s', autoflush, autocommit, expire_on_commit)

    #: A SQLAlchemy session maker
    session_maker = sessionmaker(
        bind=engine,
        autoflush=autoflush,
        autocommit=autocommit,
        expire_on_commit=expire_on_commit,
    )

    #: A SQLAlchemy session object
    session = scoped_session(
        session_maker,
        scopefunc=scopefunc
    )

    return engine, session



# -*- coding: utf-8 -*-

"""Bio2BEL is a project aimed at integrating biological databases and other structured knowledge sources.

Because they come from a wide variety of schemata, this package provides tools for converting them into the unified
framework of `Biological Expression Language <http://http://openbel.org/>`_.

This package provides guidelines, tutorials, and tools for making standardized ``bio2bel`` packages as well as a
unifying framework for integrating them.

Installation
------------
Easiest
~~~~~~~
Download the latest stable code from `PyPI <https://pypi.org/bio2bel>`_ with:

.. code-block:: sh

   $ python3 -m pip install bio2bel

Get the Latest
~~~~~~~~~~~~~~~
Download the most recent code from `GitHub <https://github.com/bio2bel/bio2bel>`_ with:

.. code-block:: sh

   $ python3 -m pip install git+https://github.com/bio2bel/bio2bel.git

For Developers
~~~~~~~~~~~~~~
Clone the repository from `GitHub <https://github.com/bio2bel/bio2bel>`_ and install in editable mode with:

.. code-block:: sh

   $ git clone https://github.com/bio2bel/bio2bel.git
   $ cd bio2bel
   $ python3 -m pip install -e .


Testing
-------
Bio2BEL is tested with Python3 on Linux using `Travis CI <https://travis-ci.org/bio2bel/bio2bel>`_.
"""

from .downloading import make_df_getter, make_downloader  # noqa: F401
from .manager import abstract_manager  # noqa: F401
from .manager.abstract_manager import AbstractManager  # noqa: F401
from .utils import get_data_dir, get_version  # noqa: F401

__version__ = '0.1.1'

__title__ = 'bio2bel'
__description__ = "A package for converting biological stuff to BEL"
__url__ = 'https://github.com/bio2bel/bio2bel'

__author__ = 'Charles Tapley Hoyt'
__email__ = 'charles.hoyt@scai.fraunhofer.de'

__license__ = 'Apache 2.0 License'
__copyright__ = 'Copyright (c) 2017-2018 Charles Tapley Hoyt'



# -*- coding: utf-8 -*-

"""Bio2BEL database models."""

import datetime
import logging

from bio2bel.constants import get_global_connection
from sqlalchemy import Column, create_engine, DateTime, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

log = logging.getLogger(__name__)

Base = declarative_base()

TABLE_PREFIX = 'bio2bel'
ACTION_TABLE_NAME = '{}_action'.format(TABLE_PREFIX)


def create_all(engine, checkfirst=True):
    """Create the tables for Bio2BEL."""
    Base.metadata.create_all(engine, checkfirst=checkfirst)


def _make_session(connection=None):
    """Make a session.

    :type connection: Optional[str]
    :rtype: sqlalchemy.orm.Session
    """
    if connection is None:
        connection = get_global_connection()

    engine = create_engine(connection)

    create_all(engine)

    session_cls = sessionmaker(bind=engine)
    session = session_cls()

    return session


def _store_helper(make_method, resource, session=None):
    """Help store an action.

    :param make_method: Either :meth:`Action.make_populate` or :meth:`Action.make_drop`
    :param str resource: The lowercase name of the resource. Ex: 'interpro'
    :param Optional[sqlalchemy.orm.Session] session: A pre-built session
    :rtype: Action
    """
    if session is None:
        session = _make_session()

    model = make_method(resource)
    session.add(model)
    session.commit()
    session.close()

    return model


class Action(Base):
    """Represents an update, dropping, population, etc. to the database."""

    __tablename__ = ACTION_TABLE_NAME

    id = Column(Integer, primary_key=True)

    resource = Column(String(32), nullable=False,
                      doc='The normalized name of the Bio2BEL package (e.g., hgnc, chebi, etc)')
    action = Column(String(32), nullable=False)
    created = Column(DateTime, nullable=False, default=datetime.datetime.utcnow, doc='The date and time of upload')

    def __repr__(self):  # noqa: D105
        return '{} {} at {}'.format(self.resource, self.action, self.created)

    @staticmethod
    def make_populate(resource):
        """Make a ``populate`` instance of :class:`Action`.

        :rtype: Action
        """
        return Action(resource=resource.lower(), action='populate')

    @staticmethod
    def make_drop(resource):
        """Make a ``drop`` instance of :class:`Action`.

        :rtype: Action
        """
        return Action(resource=resource.lower(), action='drop')

    @classmethod
    def store_populate(cls, resource, session=None):
        """Store a populate event.

        :param str resource: The normalized name of the resource to store
        :param Optional[sqlalchemy.orm.Session] session: A pre-built session
        :rtype: Action

        Example:

        >>> from bio2bel.models import Action
        >>> Action.store_populate('hgnc')
        """
        return _store_helper(cls.make_populate, resource, session=session)

    @classmethod
    def store_drop(cls, resource, session=None):
        """Store a drop event.

        :param str resource: The normalized name of the resource to store
        :param Optional[sqlalchemy.orm.Session] session: A pre-built session
        :rtype: Action

        Example:

        >>> from bio2bel.models import Action
        >>> Action.store_drop('hgnc')
        """
        return _store_helper(cls.make_drop, resource, session=session)

    @classmethod
    def ls(cls, session=None):
        """Get all actions.

        :param Optional[sqlalchemy.orm.Session] session: A pre-built session
        :rtype: list[Action]
        """
        if session is None:
            session = _make_session()

        actions = session.query(cls).order_by(cls.created.desc()).all()
        session.close()
        return actions

    @classmethod
    def count(cls, session=None):
        """Count all actions.

        :param Optional[sqlalchemy.orm.Session] session: A pre-built session
        :rtype: int
        """
        if session is None:
            session = _make_session()

        count = session.query(cls).count()
        session.close()
        return count



# -*- coding: utf-8 -*-

"""Run Bio2BEL with WSGI."""

from bio2bel.web.application import create_application

if __name__ == '__main__':
    app = create_application()
    app.run(host='0.0.0.0', port=5000, debug=True)



# -*- coding: utf-8 -*-

"""Provides abstractions over the management of SQLAlchemy connections and sessions."""

from abc import ABCMeta, abstractmethod
from functools import wraps
import logging
import os
import sys

import click

from .cli_manager import CliMixin
from .connection_manager import ConnectionManager
from ..utils import get_data_dir

__all__ = [
    'AbstractManager',
]

log = logging.getLogger(__name__)


class AbstractManagerMeta(ABCMeta):
    """Crazy metaclass to hack in a decorator to the populate function."""

    def __new__(mcs, name, bases, namespace, **kwargs):  # noqa: N804
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        cls._populate_original = cls.populate

        @wraps(cls._populate_original)
        def populate_wrapped(self, *populate_args, **populate_kwargs):
            """Populate the database."""
            cls._populate_original(self, *populate_args, **populate_kwargs)

            # Hack in the action storage
            self._store_populate()

        cls.populate = populate_wrapped

        return cls


class AbstractManager(ConnectionManager, CliMixin, metaclass=AbstractManagerMeta):
    """This is a base class for implementing your own Bio2BEL manager.

    It already includes functions to handle configuration, construction of a connection to a database using SQLAlchemy,
    creation of the tables defined by your own :func:`sqlalchemy.ext.declarative.declarative_base`, and has hooks to
    override that populate and make simple queries to the database. Since :class:`AbstractManager` inherits from
    :class:`abc.ABC` and is therefore an abstract class, there are a few class variables, functions, and properties
    that need to be overridden.

    **Overriding the Module Name**

    First, the class-level variable ``module_name`` must be set to a string corresponding to the name of the data
    source.

    .. code-block:: python

        from bio2bel import AbstractManager

        class Manager(AbstractManager):
            module_name = 'mirtarbase'  # note: use lower case module names

    In general, this should also correspond to the same value as ``MODULE_NAME`` set in ``constants.py`` and can also
    be set with an assignment to this value

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME

        class Manager(AbstractManager):
            module_name = MODULE_NAME

    **Setting the Declarative Base**

    Building on the previous example, the (private) abstract property :data:`bio2bel.AbstractManager._base` must be
    overridden to return the value from your :func:`sqlalchemy.ext.declarative.declarative_base`. We chose to make this
    an instance-level property instead of a class-level variable so each manager could have its own information about
    connections to the database.

    As a minimal example:

    .. code-block:: python

        from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base

        from bio2bel import AbstractManager

        Base: DeclarativeMeta = declarative_base()

        class Manager(AbstractManager):
            module_name = 'mirtarbase'  # note: use lower case module names

            @property
            def _base(self) -> DeclarativeMeta:
                return Base


    In general, the models should be defined in a module called ``models.py`` so the ``Base`` can also be imported.

    .. code-block:: python

        from sqlalchemy.ext.declarative import DeclarativeMeta

        from bio2bel import AbstractManager

        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self) -> DeclarativeMeta:
                return Base

    **Populating the Database**

    Deciding how to populate the database using your SQLAlchemy models is incredibly creative and can't be given a good
    example without checking real code. See the previously mentioned `implementation of a Manager
    <https://github.com/bio2bel/mirtarbase/blob/master/src/bio2bel_mirtarbase/manager.py>`_.

    .. code-block:: python

        from sqlalchemy.ext.declarative import DeclarativeMeta

        from bio2bel import AbstractManager

        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self) -> DeclarativeMeta:
                return Base

            def populate(self) -> None:
                ...

    **Checking the Database is Populated**

    A method for checking if the database has been populated already must be implemented as well. The easiest way to
    implement this is to check that there's a non-zero count of whatever the most important model in the database is.

    .. code-block:: python

        from sqlalchemy.ext.declarative import DeclarativeMeta

        from bio2bel import AbstractManager

        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self) -> DeclarativeMeta:
                return Base

            def populate(self) -> None:
                ...

            def is_populated(self) -> bool:
                return 0 < self.session.query(MyImportantModel).count()

    There are several mixins that can be optionally inherited:

    1. :py:class:`bio2bel.manager.flask_manager.FlaskMixin`: the Flask Mixin creates a Flask-Admin web application.
    2. :py:class:`bio2bel.manager.namespace_manager.BELNamespaceManagerMixin`: the BEL Namespace Manager Mixin exports
       a BEL namespace and interact with PyBEL.
    3. :py:class:`bio2bel.manager.bel_manager.BELManagerMixin`: the BEL Manager Mixin exports a BEL script
       and interact with PyBEL.
    """

    @property
    @abstractmethod
    def _base(self):
        """Return the declarative base.

        It is usually sufficient to return an instance that is module-level.

        :rtype: sqlalchemy.ext.declarative.api.DeclarativeMeta

        How to build an instance of :class:`sqlalchemy.ext.declarative.api.DeclarativeMeta` by using
        :func:`sqlalchemy.ext.declarative.declarative_base`:

        >>> from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
        >>> Base: DeclarativeMeta = declarative_base()

        Then just override this abstract property like:

        >>> @property
        >>> def _base(self) -> DeclarativeMeta:
        >>>     return Base

        Note that this property could effectively also be a static method.
        """

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(*args, **kwargs)
        self.create_all()

    @abstractmethod
    def is_populated(self):
        """Check if the database is already populated.

        :rtype: bool
        """

    @abstractmethod
    def populate(self, *args, **kwargs):
        """Populate the database."""

    @abstractmethod
    def summarize(self):
        """Summarize the database.

        :rtype: dict[str,int]
        """

    @property
    def _metadata(self):
        """Return the metadata object associated with this manager's declarative base."""
        return self._base.metadata

    def create_all(self, check_first=True):
        """Create the empty database (tables).

        :param bool check_first: Defaults to True, don't issue CREATEs for tables already present
         in the target database. Defers to :meth:`sqlalchemy.sql.schema.MetaData.create_all`
        """
        self._metadata.create_all(self.engine, checkfirst=check_first)

    def drop_all(self, check_first=True):
        """Drop all tables from the database.

        :param bool check_first: Defaults to True, only issue DROPs for tables confirmed to be
          present in the target database. Defers to :meth:`sqlalchemy.sql.schema.MetaData.drop_all`
        """
        self._metadata.drop_all(self.engine, checkfirst=check_first)
        self._store_drop()

    def _get_query(self, model):
        """Get a query for the given model using this manager's session.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :return: a SQLAlchemy query
        """
        return self.session.query(model)

    def _count_model(self, model):
        """Count the number of the given model in the database.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :rtype: int
        """
        return self._get_query(model).count()

    def _list_model(self, model):
        """Get all instances of the given model in the database.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :rtype: list
        """
        return self._get_query(model).all()

    @staticmethod
    def _cli_add_populate(main):
        """Add the populate command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_populate(main)

    @staticmethod
    def _cli_add_drop(main):
        """Add the drop command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_drop(main)

    @staticmethod
    def _cli_add_cache(main):
        """Add the cache command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_cache(main)

    @staticmethod
    def _cli_add_summarize(main):
        """Add the summarize command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_summarize(main)

    @classmethod
    def get_cli(cls):
        """Get the :mod:`click` main function to use as a command line interface.

        :rtype: click.Group
        """
        main = super().get_cli()

        cls._cli_add_populate(main)
        cls._cli_add_drop(main)
        cls._cli_add_cache(main)
        cls._cli_add_summarize(main)

        return main


def add_cli_populate(main):
    """Add a ``populate`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('--reset', is_flag=True, help='Nuke database first')
    @click.option('--force', is_flag=True, help='Force overwrite if already populated')
    @click.pass_obj
    def populate(manager, reset, force):
        """Populate the database."""
        if reset:
            click.echo('Deleting the previous instance of the database')
            manager.drop_all()
            click.echo('Creating new models')
            manager.create_all()

        if manager.is_populated() and not force:
            click.echo('Database already populated. Use --force to overwrite')
            sys.exit(0)

        manager.populate()

    return main


def add_cli_drop(main):
    """Add a ``drop`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-y', '--yes', is_flag=True)
    @click.pass_obj
    def drop(manager, yes):
        """Drop the database."""
        if yes or click.confirm('Drop everything?'):
            manager.drop_all()

    return main


def add_cli_cache(main):
    """Add several commands to main :mod:`click` function for handling the cache.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.group()
    def cache():
        """Manage cached data."""

    @cache.command()
    @click.pass_obj
    def ls(manager):
        """List files in the cache."""
        data_dir = get_data_dir(manager.module_name)

        for path in os.listdir(data_dir):
            click.echo(path)

    @cache.command()
    @click.pass_obj
    def clear(manager):
        """Clear all files from the cache."""
        data_dir = get_data_dir(manager.module_name)

        for path in os.listdir(data_dir):
            if path in {'config.ini', 'cache.db'}:
                continue
            os.remove(os.path.join(data_dir, path))

    return main


def add_cli_summarize(main):
    """Add a ``summarize`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.pass_obj
    def summarize(manager):
        """Summarize the contents of the database."""
        for name, count in sorted(manager.summarize().items()):
            click.echo('{}: {}'.format(name.capitalize(), count))

    return main



# -*- coding: utf-8 -*-

"""Provide abstractions over BEL generation procedures."""

from abc import ABC, abstractmethod
import logging
import sys

import click

from .cli_manager import CliMixin

log = logging.getLogger(__name__)

__all__ = [
    'BELManagerMixin',
]


class BELManagerMixin(ABC, CliMixin):
    """A mixin for generating a :class:`pybel.BELGraph` representing BEL.

    First, you'll have to make sure that :mod:`pybel` is installed. This can be done with pip like:

    .. code-block:: bash

        $ pip install pybel

    To use this mixin, you need to properly implement the AbstractManager, and additionally define a function
    named ``to_bel`` that returns a BEL graph.

    .. code-block:: python

        >>> from bio2bel import AbstractManager
        >>> from bio2bel.manager.bel_manager import BELManagerMixin
        >>> from pybel import BELGraph
        >>>
        >>> class MyManager(AbstractManager, BELManagerMixin):
        ...     def to_bel(self) -> BELGraph:
        ...         pass
    """

    @abstractmethod
    def to_bel(self, *args, **kwargs):
        """Convert the database to BEL.

        :rtype: pybel.BELGraph

        Example implementation outline:

        .. code-block:: python

            from bio2bel import AbstractManager
            from bio2bel.manager.bel_manager import BELManagerMixin
            from pybel import BELGraph
            from .models import Interaction

            class MyManager(AbstractManager, BELManagerMixin):
                module_name = 'mirtarbase'
                def to_bel(self):
                    rv = BELGraph(
                        name='miRTarBase',
                        version='1.0.0',
                    )

                    for interaction in self.session.query(Interaction):
                        mirna = mirna_dsl('mirtarbase', interaction.mirna.mirtarbase_id)
                        rna = rna_dsl('hgnc', interaction.target.hgnc_id)

                        rv.add_qualified_edge(
                            mirna,
                            rna,
                            DECREASES,
                            ...
                        )

                    return rv
        """

    @staticmethod
    def _cli_add_to_bel(main):
        """Add the export BEL command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_to_bel(main)

    @staticmethod
    def _cli_add_upload_bel(main):
        """Add the upload BEL command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_upload_bel(main)

    @classmethod
    def get_cli(cls):
        """Get a :mod:`click` main function with added BEL commands.

        :rtype: click.Group
        """
        main = super().get_cli()

        @main.group()
        def bel():
            """Manage BEL."""

        cls._cli_add_to_bel(bel)
        cls._cli_add_upload_bel(bel)

        return main


def add_cli_to_bel(main):
    """Add several command to main :mod:`click` function related to export to BEL.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-o', '--output', type=click.File('w'), default=sys.stdout)
    @click.pass_obj
    def write(manager, output):
        """Write as BEL Script."""
        from pybel import to_bel
        graph = manager.to_bel()
        to_bel(graph, output)


def add_cli_upload_bel(main):
    """Add several command to main :mod:`click` function related to export to BEL.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-c', '--connection')
    @click.pass_obj
    def upload(manager, connection):
        """Upload BEL to network store."""
        import pybel
        graph = manager.to_bel()
        pybel_manager = pybel.Manager(connection=connection)
        pybel.to_database(graph, manager=pybel_manager)

    return main



# -*- coding: utf-8 -*-

"""Tests for the Bio2BEL AbstractManager."""

import unittest

from bio2bel import AbstractManager
from bio2bel.exc import Bio2BELMissingNameError, Bio2BELModuleCaseError
from bio2bel.models import Action
from bio2bel.testing import AbstractTemporaryCacheClassMixin, MockConnectionMixin, TemporaryConnectionMethodMixin
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
import tests.constants
from tests.constants import NUMBER_TEST_MODELS


class TestManagerFailures(TemporaryConnectionMethodMixin):
    """Test improperly implement AbstractManager."""

    def test_missing_all_abstract(self):
        """Test that the abstract class can't be instantiated."""
        class Manager(AbstractManager):
            """An incompletely implement AbstractManager."""

        with self.assertRaises(TypeError):  # cant's instantiate abstract class
            Manager(self.connection)

    def test_fail_instantiation_2(self):
        """Test that the abstract class can't be instantiated."""
        base = declarative_base()

        class Manager(AbstractManager):
            """An incompletely implement AbstractManager."""

            @property
            def _base(self):
                return base

        with self.assertRaises(TypeError):
            Manager(connection=self.connection)

    def test_fail_instantiation_3(self):
        """Test that the abstract class can't be instantiated."""
        class Manager(AbstractManager):
            """An incompletely implement AbstractManager."""

            def populate(self):
                """Populate the database."""

        with self.assertRaises(TypeError):
            Manager(connection=self.connection)

    def test_undefined_module_name(self):
        """Test error thrown if module name isn't set."""
        base = declarative_base()

        class Manager(AbstractManager):
            """An improperly implemented AbstractManager that is missing the module_name class variable."""

            @property
            def _base(self):
                return base

            def is_populated(self):
                """Check if the database is already populated."""
                pass

            def populate(self):
                """Populate the database."""
                pass

            def summarize(self):
                """Summarize the database."""
                pass

        with self.assertRaises(Bio2BELMissingNameError):
            Manager(connection=self.connection)

    def test_module_name_case(self):
        """Test error thrown if module name is weird case."""
        base = declarative_base()

        class Manager(AbstractManager):
            """A test manager that checks the module name is lower cased."""

            module_name = 'TESTOMG'

            @property
            def _base(self):
                return base

            def is_populated(self):
                """Check if the database is already populated."""
                pass

            def populate(self):
                """Populate the database."""
                pass

            def summarize(self):
                """Summarize the database."""
                pass

        with self.assertRaises(Bio2BELModuleCaseError):
            Manager(connection=self.connection)


class TestConnectionDropping(MockConnectionMixin, AbstractTemporaryCacheClassMixin):
    """Tests dropping the database."""

    Manager = tests.constants.Manager

    def test_no_exist(self):
        """Check if the database gets dropped that stuff breaks."""
        with self.mock_global_connection:  # don't want to worry about that drop_app hook
            self.assertEqual(0, Action.count())
            self.manager.drop_all()
            self.assertEqual(1, Action.count())

        with self.assertRaises(OperationalError):
            self.manager.get_model_by_model_id(5)


class TestConnectionLoading(AbstractTemporaryCacheClassMixin):
    """Tests the connection is loaded properly."""

    Manager = tests.constants.Manager

    def test_connection(self):
        """Test the type of the connection."""
        self.assertIsNotNone(self.connection)
        self.assertIsInstance(self.connection, str)

    def test_manager_passes(self):
        """Test the connection is inside the manager properly."""
        self.assertEqual(self.connection, str(self.manager.engine.url))

    def test_repr(self):
        """Test the repr function of the AbstractManager."""
        self.assertEqual('<TestManager url={}>'.format(self.connection), repr(self.manager))

    def test_get_missing_model(self):
        """Test population."""
        self.assertFalse(self.manager.is_populated())
        self.manager.populate()
        self.assertTrue(self.manager.is_populated())

        self.assertEqual(NUMBER_TEST_MODELS, self.manager.count_model())

        self.assertIsNone(self.manager.get_model_by_model_id(0))
        self.assertIsNone(self.manager.get_model_by_model_id(1))
        self.assertIsNone(self.manager.get_model_by_model_id(2))
        self.assertIsNone(self.manager.get_model_by_model_id(3))
        self.assertIsNone(self.manager.get_model_by_model_id(4))
        self.assertIsNone(self.manager.get_model_by_model_id(150))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Import hook for extensions."""

import logging
import sys

from pkg_resources import iter_entry_points

__all__ = [
    'ExtensionImporter',
]

log = logging.getLogger(__name__)


class ExtensionImporter(object):
    """An importer and loader for package resources registered under a particular group."""

    def __init__(self, group):
        """Initialize the importer with the group name.

        :param str group: a string representing the package resources entry_points group that will be used
        """
        self.group = group

    @property
    def _group_with_dot(self):
        return '{}.'.format(self.group)

    def install(self):
        """Call this method to install the new importer to :code:`sys.meta_path`. Should probably only be done once."""
        sys.meta_path.append(self)

    def find_module(self, fullname, path=None):
        """Find a module if its name starts with :code:`self.group` and is registered."""
        if not fullname.startswith(self._group_with_dot):
            return
        end_name = fullname[len(self._group_with_dot):]
        for entry_point in iter_entry_points(group=self.group, name=None):
            if entry_point.name == end_name:
                return self

    def load_module(self, fullname):
        """Load a module if its name starts with :code:`self.group` and is registered."""
        if fullname in sys.modules:
            return sys.modules[fullname]
        end_name = fullname[len(self._group_with_dot):]
        for entry_point in iter_entry_points(group=self.group, name=end_name):
            mod = entry_point.load()
            sys.modules[fullname] = mod
            return mod



# -*- coding: utf-8 -*-

"""The Bio2BEL web application."""

import importlib
import logging

from bio2bel.constants import DEFAULT_CACHE_CONNECTION
from bio2bel.manager.connection_manager import build_engine_session
import flask
from flask_admin import Admin
import flask_bootstrap
from pkg_resources import iter_entry_points, VersionConflict

log = logging.getLogger(__name__)

ui = flask.Blueprint('ui', __name__)

web_modules = {}
add_admins = {}

for entry_point in iter_entry_points(group='bio2bel', name=None):
    entry = entry_point.name

    try:
        bio2bel_module = entry_point.load()
    except VersionConflict:
        log.warning('Version conflict in %s', entry)
        continue

    try:
        web_modules[entry] = bio2bel_module.web
    except AttributeError:
        try:
            web_modules[entry] = importlib.import_module('bio2bel_{}.web'.format(entry))
        except ImportError:
            log.warning('no submodule bio2bel_%s.web', entry)
            continue

    try:
        add_admins[entry] = web_modules[entry].add_admin
    except AttributeError:
        log.warning('no function bio2bel_%s.web.add_admin', entry)
        continue


@ui.route('/')
def home():
    """Show the home page."""
    return flask.render_template('index.html', entries=sorted(add_admins))


def create_application(connection=None):
    """Create a Flask application.

    :param Optional[str] connection: A connection string
    :rtype: flask.Flask
    """
    app = flask.Flask(__name__)

    flask_bootstrap.Bootstrap(app)
    Admin(app)

    connection = connection or DEFAULT_CACHE_CONNECTION
    engine, session = build_engine_session(connection)

    for name, add_admin in add_admins.items():
        url = '/{}'.format(name)
        add_admin(app, session, url=url, endpoint=name, name=name)
        log.debug('added %s - %s to %s', name, add_admin, url)

    app.register_blueprint(ui)

    return app



# -*- coding: utf-8 -*-

"""Tests the Flask web application generation utilities."""

from bio2bel.exc import Bio2BELMissingModelsError
from bio2bel.manager.flask_manager import FlaskMixin
from bio2bel.testing import TemporaryConnectionMethodMixin
from flask_admin.contrib.sqla import ModelView
from tests.constants import Manager, Model


class WrongFlaskTestManager(FlaskMixin):
    """An implementation of an AbstractManager that is unable to produce a Flask app."""

    module_name = 'test'


class FlaskTestManager(FlaskMixin, Manager):
    """Extends the test Manager for generating a Flask application."""

    flask_admin_models = [Model]


class TruncatedModelView(ModelView):
    """A truncated Flask Admin view."""

    column_exclude_list = ['model_id']


class FlaskTestViewManager(Manager, FlaskMixin):
    """Extends the test Manager for generating a Flask application."""

    flask_admin_models = [(Model, TruncatedModelView)]


class FlaskFailureTestViewManager(Manager, FlaskMixin):
    """Extends the test Manager for generating a Flask application."""

    flask_admin_models = [(Model, TruncatedModelView, 'junk!')]


class TestFlask(TemporaryConnectionMethodMixin):
    """Tests Flask application generation."""

    def test_missing_models(self):
        """Test exceptions are thrown properly for an improperly implemented AbstractManager."""
        self.assertIs(WrongFlaskTestManager.flask_admin_models, ...)

        with self.assertRaises(Bio2BELMissingModelsError):
            WrongFlaskTestManager(connection=self.connection)

    def test_app(self):
        """Test the successful generation of a flask application."""
        manager = FlaskTestManager(connection=self.connection)

        self.assertFalse(manager.is_populated())
        manager.populate()
        self.assertTrue(manager.is_populated())

        app = manager.get_flask_admin_app()
        client = app.test_client()

        home_rv = client.get('/')
        self.assertIn(Model.__name__.encode('utf-8'), home_rv.data)

        rv = client.get('/{}'.format(Model.__name__.lower()), follow_redirects=True)
        self.assertIn(b'MODEL:1', rv.data)

    def test_app_truncated_view(self):
        """Test the ability to define tuple views."""
        manager = FlaskTestViewManager(connection=self.connection)

        self.assertFalse(manager.is_populated())
        manager.populate()
        self.assertTrue(manager.is_populated())

        app = manager.get_flask_admin_app()
        client = app.test_client()

        home_rv = client.get('/')
        self.assertIn(Model.__name__.encode('utf-8'), home_rv.data)

        rv = client.get('/{}'.format(Model.__name__.lower()), follow_redirects=True)
        self.assertNotIn(b'MODEL:1', rv.data)
        self.assertIn(b'1111', rv.data)

    def test_app_view_failure(self):
        """Test the ability to define tuple views."""
        manager = FlaskFailureTestViewManager(connection=self.connection)
        manager.populate()

        with self.assertRaises(TypeError):
            manager.get_flask_admin_app()



# -*- coding: utf-8 -*-

"""Provides abstractions over the generation of a command line interface."""

import logging

import click

from ..utils import get_version

__all__ = [
    'CliMixin',
]


class CliMixin:
    """A mixin for building a CLI.

    Must be used as a mixin for a subclass of :class:`bio2bel.manager.connection_manager.ConnectionManager`.
    """

    @classmethod
    def get_cli(cls):
        """Build a :mod:`click` CLI main function.

        :param Type[AbstractManager] cls: A Manager class
        :return: The main function for click
        :rtype: click.core.Group
        """
        group_help = 'Default connection at {}\n\nusing Bio2BEL v{}'.format(cls._get_connection(), get_version())

        @click.group(help=group_help)
        @click.option('-c', '--connection', default=cls._get_connection(),
                      help='Defaults to {}'.format(cls._get_connection()))
        @click.pass_context
        def main(ctx, connection):
            """Bio2BEL CLI."""
            logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            logging.getLogger('bio2bel.utils').setLevel(logging.WARNING)
            ctx.obj = cls(connection=connection)

        return main



# -*- coding: utf-8 -*-

"""Testing constants and utilities for Bio2BEL."""

import logging

from bio2bel.manager.namespace_manager import BELNamespaceManagerMixin, Bio2BELMissingNamespaceModelError
from bio2bel.testing import AbstractTemporaryCacheMethodMixin, MockConnectionMixin, TemporaryConnectionMethodMixin
from click.testing import CliRunner
import pybel
from pybel.manager.models import Base as PyBELBase, Namespace, NamespaceEntry
from tests.constants import Manager, Model, NUMBER_TEST_MODELS, TEST_MODEL_ID_FORMAT, TEST_MODEL_NAME_FORMAT

log = logging.getLogger(__name__)


class NamespaceManager(Manager, BELNamespaceManagerMixin):
    """Use parts of the test manager and finish the abstract namespace manager."""

    namespace_model = Model

    # automate by defining identifier column?

    def _create_namespace_entry_from_model(self, model, namespace=None):
        return NamespaceEntry(name=model.name, identifier=model.model_id, namespace=namespace)

    def _get_identifier(self, model):
        return model.model_id


class TestFailure(TemporaryConnectionMethodMixin):
    """Test various failures and exceptions."""

    def test_type_failure(self):
        """Test that a manager with implemented functions, but no class variable, fails to instantiate."""
        class _TestManager(Manager, BELNamespaceManagerMixin):
            """A manager that has functions implemented, but no class variables."""

            def _create_namespace_entry_from_model(self, model, namespace=None):
                """Create a namespace entry."""
                return NamespaceEntry(name=model.name, identifier=model.model_id, namespace=namespace)

            def _get_identifier(self, model):
                """Get the identifier from a model."""
                return model.model_id

        with self.assertRaises(Bio2BELMissingNamespaceModelError):
            _TestManager(connection=self.connection)

    def test_instantiation_failure(self):
        """Test that a manager without implementation fails to instantiate."""
        class _TestManager(Manager, BELNamespaceManagerMixin):
            """Use parts of the test manager and finish the abstract namespace manager."""

            namespace_model = Model

        with self.assertRaises(TypeError):
            _TestManager(connection=self.connection)


class TestAwesome(AbstractTemporaryCacheMethodMixin):
    """Tests for namespace management."""

    Manager = NamespaceManager

    def setUp(self):
        """Set up the class with a PyBEL database."""
        super().setUp()
        PyBELBase.metadata.create_all(self.manager.engine, checkfirst=True)

    def populate(self):
        """Populate the manager."""
        self.manager.populate()

    def test_namespace_name(self):
        """Test the name generated by the manager."""
        self.assertEqual('test', self.manager.module_name)  # this is defined in the tests
        self.assertEqual('TEST', self.manager._get_namespace_keyword())
        self.assertEqual('_TEST', self.manager._get_namespace_url())

    def test_make_namespace(self):
        """Test when a namespace must be generated."""
        namespace = self.manager._make_namespace()
        self.assertIsNotNone(namespace)
        self.assertIsInstance(namespace, Namespace)
        self.assertEqual('TEST', namespace.keyword)
        self.assertEqual('_TEST', namespace.url)

        self.assertEqual(NUMBER_TEST_MODELS, namespace.entries.count())

        em = {
            entry.identifier: entry
            for entry in namespace.entries
        }

        for i in range(NUMBER_TEST_MODELS):
            model_id = TEST_MODEL_ID_FORMAT.format(i)

            self.assertIn(model_id, em)
            model = em[model_id]

            self.assertEqual(model_id, model.identifier)
            self.assertEqual(TEST_MODEL_NAME_FORMAT.format(i), model.name)
            self.assertEqual(namespace, model.namespace)

        # TODO fix cascade on namespace to namespace entries
        # self.manager.clear_bel_namespace()
        # self.assertIsNone(self.manager._get_default_namespace())
        # self.assertEqual(0, self.manager.session.query(Namespace).count())
        # self.assertEqual(0, self.manager.session.query(NamespaceEntry).count())

    def test_update_namespace(self):
        """Test when a namespace must be updated."""
        namespace = self.manager._make_namespace()

        # mock some sort of changes to the database

        _number_to_add = 4

        models = [
            Model.from_id(model_id)
            for model_id in range(NUMBER_TEST_MODELS + 1, NUMBER_TEST_MODELS + 1 + _number_to_add)
        ]
        self.manager.session.add_all(models)
        self.manager.session.commit()

        self.manager._update_namespace(namespace)
        self.assertEqual(NUMBER_TEST_MODELS + _number_to_add, namespace.entries.count())


class TestCli(MockConnectionMixin):
    """Tests the CLI for uploading a BEL namespace."""

    def setUp(self):
        """Set up a CliRunner and an accompanying CLI for each test."""
        self.runner = CliRunner()
        self.main = NamespaceManager.get_cli()
        self.manager = Manager(connection=self.connection)
        self.manager.populate()

    def test_to_bel_namespace(self):
        """Test the population function can be run."""
        self.assertEqual(5, self.manager.count_model(), msg='manager should be populated')

        pybel_manager = pybel.Manager(connection=self.connection)
        self.assertEqual(0, pybel_manager.count_namespaces())

        args = [
            '--connection',
            self.connection,
            'belns',
            'upload',
        ]
        self.runner.invoke(self.main, args)

        self.assertEqual(1, pybel_manager.count_namespaces())



# -*- coding: utf-8 -*-

"""Testing utilities for Bio2BEL.

This module has tools for quickly writing unit tests with :mod:`unittest` that involve the usage of a mock data with
a Bio2BEL manager.
"""

import logging
import os
import tempfile
import unittest
from unittest import mock

from .exc import Bio2BELManagerTypeError, Bio2BELTestMissingManagerError
from .manager.abstract_manager import AbstractManager

__all__ = [
    'TemporaryConnectionMethodMixin',
    'TemporaryConnectionMixin',
    'MockConnectionMixin',
    'AbstractTemporaryCacheMethodMixin',
    'AbstractTemporaryCacheClassMixin',
    'make_temporary_cache_class_mixin',
]

log = logging.getLogger(__name__)


class TemporaryConnectionMethodMixin(unittest.TestCase):
    """Creates a :class:`unittest.TestCase` that has a persistent file for use with SQLite during testing."""

    def setUp(self):
        """Create a temporary file to use as a persistent database throughout tests in this class."""
        super().setUp()

        self.fd, self.path = tempfile.mkstemp()
        self.connection = 'sqlite:///' + self.path
        log.info('test database at %s', self.connection)

    def tearDown(self):
        """Close the connection to the database and removes the files created for it."""
        os.close(self.fd)
        os.remove(self.path)


class TemporaryConnectionMixin(unittest.TestCase):
    """Creates a :class:`unittest.TestCase` that has a persistent file for use with SQLite during testing."""

    fd, path = None, None
    connection = None

    @classmethod
    def setUpClass(cls):
        """Create a temporary file to use as a persistent database throughout tests in this class.

        Subclasses of :class:`TemporaryCacheClsMixin` can extend :func:`TemporaryCacheClsMixin.setUpClass` to populate
        the database.
        """
        super().setUpClass()

        cls.fd, cls.path = tempfile.mkstemp()
        cls.connection = 'sqlite:///' + cls.path
        log.info('test database at %s', cls.connection)

    @classmethod
    def tearDownClass(cls):
        """Close the connection to the database and removes the files created for it."""
        os.close(cls.fd)
        os.remove(cls.path)


class MockConnectionMixin(TemporaryConnectionMixin):
    """Allows for testing with a consistent connection without changing the configuration."""

    def setUp(self):
        """Set up the test with a mock connection string.

        Add two class-level variables: ``mock_global_connection`` and ``mock_module_connection`` that can be
        used as context managers to mock the bio2bel connection getter functions.
        """
        super().setUp()

        def mock_connection():
            """Get the connection enclosed by this class.

            :rtype: str
            """
            return self.connection

        self.mock_global_connection = mock.patch('bio2bel.models.get_global_connection', mock_connection)
        self.mock_module_connection = mock.patch('bio2bel.utils.get_connection', mock_connection)


class AbstractTemporaryCacheMethodMixin(TemporaryConnectionMethodMixin):
    """Allows for testing with a consistent connection and creation of a manager class wrapping that connection.

    Requires the class variable ``Manager`` to be overriden with the class corresponding to the manager to be used that
    is a subclass of :class:`bio2bel.AbstractManager`.
    """

    Manager = ...
    manager = None

    def setUp(self):
        """Set up the class with the given manager and allows an optional populate hook to be overridden."""
        if self.Manager is ...:
            raise Bio2BELTestMissingManagerError('Must override class variable "Manager" with subclass of '
                                                 'bio2bel.AbstractManager')

        if not issubclass(self.Manager, AbstractManager):
            raise Bio2BELManagerTypeError('Manager must be a subclass of bio2bel.AbstractManager')

        super().setUp()

        self.manager = self.Manager(connection=self.connection)
        self.populate()

    def tearDown(self):
        """Close the connection in the manager and deletes the temporary database."""
        self.manager.session.close()
        super().tearDown()

    def populate(self):
        """Populate the database.

        This stub should be overridden.
        """


class AbstractTemporaryCacheClassMixin(TemporaryConnectionMixin):
    """Allows for testing with a consistent connection and creation of a manager class wrapping that connection.

    Requires the class variable ``Manager`` to be overriden with the class corresponding to the manager to be used that
    is a subclass of :class:`bio2bel.AbstractManager`.
    """

    Manager = ...
    manager = None

    @classmethod
    def setUpClass(cls):
        """Set up the class with the given manager and allows an optional populate hook to be overridden."""
        if cls.Manager is ...:
            raise Bio2BELTestMissingManagerError('Must override class variable "Manager" with subclass of '
                                                 'bio2bel.AbstractManager')

        if not issubclass(cls.Manager, AbstractManager):
            raise Bio2BELManagerTypeError('Manager must be a subclass of bio2bel.AbstractManager')

        super().setUpClass()

        cls.manager = cls.Manager(connection=cls.connection)
        cls.populate()

    @classmethod
    def tearDownClass(cls):
        """Close the connection in the manager and deletes the temporary database."""
        cls.manager.session.close()
        super().tearDownClass()

    @classmethod
    def populate(cls):
        """Populate the database.

        This stub should be overridden.
        """


def make_temporary_cache_class_mixin(manager_cls):
    """Build a testing class that has a Bio2BEL manager instance ready to go.

    :param type[bio2bel.AbstractManager] manager_cls: A Bio2BEL manager
    :rtype: type[AbstractTemporaryCacheClassMixin]
    """
    class TemporaryCacheClassMixin(AbstractTemporaryCacheClassMixin):
        Manager = manager_cls

    return TemporaryCacheClassMixin



# -*- coding: utf-8 -*-

"""Test that actions are stored properly for population and dropping."""

import logging

from bio2bel.models import Action, create_all
from bio2bel.testing import MockConnectionMixin, TemporaryConnectionMethodMixin
from tests.constants import Manager

log = logging.getLogger(__name__)


class TestActions(TemporaryConnectionMethodMixin, MockConnectionMixin):
    """Test actions."""

    def test_action(self):
        """Test actions with live database."""
        manager = Manager(connection=self.connection)
        create_all(manager.engine)

        self.assertEqual(0, Action.count(session=manager.session))

        self.assertFalse(manager.is_populated())
        manager.populate(return_true=True)
        self.assertTrue(manager.is_populated())

        # just check this tiny implementation detail
        self.assertEqual(manager.last_populate_kwargs, {'return_true': True})

        self.assertEqual(1, Action.count(session=manager.session))
        actions = Action.ls(session=manager.session)
        action = actions[0]
        self.assertEqual(manager.module_name, action.resource)
        self.assertEqual('populate', action.action)

    def test_action_mocked(self):
        """Test actions with mocked database."""
        with self.mock_global_connection:
            self.assertEqual(0, Action.count())

            manager = Manager(self.connection)
            self.assertFalse(manager.is_populated())
            manager.populate()
            self.assertTrue(manager.is_populated())

            self.assertEqual(1, Action.count())
            actions = Action.ls()
            action = actions[0]
            self.assertEqual(manager.module_name, action.resource)
            self.assertEqual('populate', action.action)



# -*- coding: utf-8 -*-

"""Testing constants and utilities for Bio2BEL."""

import logging

from bio2bel.manager.abstract_manager import AbstractManager
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

log = logging.getLogger(__name__)

TestBase = declarative_base()

NUMBER_TEST_MODELS = 5
TEST_MODEL_ID_FORMAT = 'MODEL:{}'
TEST_MODEL_NAME_FORMAT = '{0}{0}{0}{0}{0}'


class Model(TestBase):
    """A test model."""

    __tablename__ = 'test_model'

    id = Column(Integer, primary_key=True)

    model_id = Column(String(15), nullable=False, index=True, unique=True)
    name = Column(String(255), nullable=False, index=True)

    @staticmethod
    def from_id(model_id):
        """Create a test Model from a given integer identifier.

        :param int model_id:
        :rtype: Model
        """
        return Model(
            model_id=TEST_MODEL_ID_FORMAT.format(model_id),
            name=TEST_MODEL_NAME_FORMAT.format(model_id),
        )


class Manager(AbstractManager):
    """Manager for running tests."""

    module_name = 'test'

    def __init__(self, *args, **kwargs):
        """Instantiate the manager."""
        super().__init__(*args, **kwargs)

        self.last_populate_args = []
        self.last_populate_kwargs = {}

    @property
    def _base(self):
        return TestBase

    def get_model_by_model_id(self, model_id):
        """Get a model if it exists by its identifier.

        :param str model_id: A Model identifier
        :rtype: Optional[Model]
        """
        return self.session.query(Model).filter(Model.model_id == model_id).one_or_none()

    def count_model(self):
        """Count the test model.

        :rtype: int
        """
        return self._count_model(Model)

    def list_model(self):
        """Get all models.

        :rtype: list[Model]
        """
        return self._list_model(Model)

    def is_populated(self):
        """Check if the database is already populated.

        :rtype: bool
        """
        return 0 < self.count_model()

    def populate(self, *args, **kwargs):
        """Add five models to the store."""
        models = [
            Model.from_id(model_id)
            for model_id in range(NUMBER_TEST_MODELS)
        ]
        self.session.add_all(models)
        self.session.commit()

        if args:
            self.last_populate_args = args
            log.critical('args: %s', args)

        if kwargs:
            self.last_populate_kwargs = kwargs
            log.critical('kwargs: %s', kwargs)

    def summarize(self):
        """Summarize the database.

        :rtype: dict[str,int]
        """
        return dict(
            models=self.count_model(),
        )



# -*- coding: utf-8 -*-

"""Tests the CLI generation utilities."""

from bio2bel.testing import MockConnectionMixin
from click.testing import CliRunner
from sqlalchemy.exc import OperationalError
from tests.constants import Manager


class TestCli(MockConnectionMixin):
    """Tests the CLI generator."""

    def setUp(self):
        """Set up a CliRunner and an accompanying CLI for each test."""
        self.runner = CliRunner()
        self.main = Manager.get_cli()

    def test_populate(self):
        """Test the population function can be run."""
        args = [
            '--connection',
            self.connection,
            'populate'
        ]
        self.runner.invoke(self.main, args)

        manager = Manager(connection=self.connection)
        self.assertEqual(5, manager.count_model())

        args = [
            '--connection',
            self.connection,
            'drop',
            '-y'
        ]
        self.runner.invoke(self.main, args)

        with self.assertRaises(OperationalError):
            manager.count_model()



# -*- coding: utf-8 -*-

"""Tests the Bio2BEL testing utilities."""

import unittest

from bio2bel.exc import Bio2BELManagerTypeError, Bio2BELTestMissingManagerError
from bio2bel.testing import (
    AbstractTemporaryCacheClassMixin, AbstractTemporaryCacheMethodMixin, make_temporary_cache_class_mixin,
)
from tests.constants import Manager


class TestMethodCacheBuild(unittest.TestCase):
    """Tests the instantiation of concrete implementation of the :class:`AbstractManager`."""

    def test_missing_manager(self):
        """Test that an incorrectly built AbstractTemporaryCacheMethodMixin won't run."""
        class TestMissingManagerMixin(AbstractTemporaryCacheMethodMixin):
            """An :class:`AbstractTemporaryCacheClassMixin` that is missing the Manager class variable."""

            def test_dummy(self):
                """Test dummy."""
                self.assertTrue(True)

        with self.assertRaises(Bio2BELTestMissingManagerError):
            m = TestMissingManagerMixin()
            m.setUp()

    def test_manager_wrong_type(self):
        """Test that an incorrectly built AbstractTemporaryCacheClassMixin won't run."""
        class RandomClass(object):
            """This class is not the right type and should throw a Bio2BELManagerTypeError."""

        class TestManagerWrongTypeMixin(AbstractTemporaryCacheMethodMixin):
            """This is a malformed :class:`AbstractTemporaryCacheClassMixin`."""

            Manager = RandomClass

            def test_dummy(self):
                """This is a dummy test."""
                self.assertTrue(True)

        with self.assertRaises(Bio2BELManagerTypeError):
            m = TestManagerWrongTypeMixin()
            m.setUp()


class TestClassCacheBuild(unittest.TestCase):
    """Tests the instantiation of concrete implementation of the :class:`AbstractManager`."""

    def test_missing_manager(self):
        """Test that an incorrectly built AbstractTemporaryCacheClassMixin won't run."""
        class TestMissingManagerMixin(AbstractTemporaryCacheClassMixin):
            """An :class:`AbstractTemporaryCacheClassMixin` that is missing the Manager class variable."""

            def test_dummy(self):
                """Test dummy."""
                self.assertTrue(True)

        with self.assertRaises(Bio2BELTestMissingManagerError):
            TestMissingManagerMixin.setUpClass()

    def test_manager_wrong_type(self):
        """Test that an incorrectly built AbstractTemporaryCacheClassMixin won't run."""
        class RandomClass(object):
            """This class is not the right type and should throw a Bio2BELManagerTypeError."""

        class TestManagerWrongTypeMixin(AbstractTemporaryCacheClassMixin):
            """This is a malformed :class:`AbstractTemporaryCacheClassMixin`."""

            Manager = RandomClass

            def test_dummy(self):
                """This is a dummy test."""
                self.assertTrue(True)

        with self.assertRaises(Bio2BELManagerTypeError):
            TestManagerWrongTypeMixin.setUpClass()


class TestTesting(make_temporary_cache_class_mixin(Manager)):
    """Tests :func:`bio2bel.testing.make_temporary_cache_class_mixin`."""

    def test_self(self):
        """Test that this test is an instance of AbstractTemporaryCacheClassMixin."""
        self.assertIsInstance(self, AbstractTemporaryCacheClassMixin)

    def test_has_attributes(self):
        """Test class has a manager instance."""
        self.assertTrue(hasattr(self, 'manager'))
        self.assertIsNotNone(self.manager)
        self.assertIsInstance(self.manager, Manager)


class TestPopulated(TestTesting):
    """Test the :func:`bio2bel.testing.make_temporary_cache_class_mixin`."""

    @classmethod
    def populate(cls):
        """Populate the database."""
        cls.manager.populate()

    def test_populated(self):
        """Test that the correct number of models have been added to the database."""
        self.assertEqual(5, self.manager.count_model())


class TestPopulatedKwargs(TestTesting):
    """Test the :func:`bio2bel.testing.make_temporary_cache_class_mixin` with kwargs in the populate function."""

    @classmethod
    def populate(cls):
        """Populate the database."""
        cls.manager.populate(return_true=True)

    def test_populated(self):
        """Test that the correct number of models have been added to the database."""
        self.assertEqual(5, self.manager.count_model())



# -*- coding: utf-8 -*-

"""This module runs the command line interface for Bio2BEL."""

import logging

from .cli import main

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    main()



# -*- coding: utf-8 -*-

"""Setup.py for Bio2BEL."""

import codecs
import os
import re

import setuptools

PACKAGES = setuptools.find_packages(where='src')
KEYWORDS = ['Biological Expression Language', 'BEL', 'Systems Biology', 'Networks Biology']
META_PATH = os.path.join('src', 'bio2bel', '__init__.py')
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering :: Bio-Informatics'
]
INSTALL_REQUIRES = [
    'sqlalchemy',
    'click',
    'pandas',
    'tqdm',
]
EXTRAS_REQUIRE = {
    'web': [
        'flask',
        'flask-admin',
    ],
    'namespaces': [
        'pybel',
    ]
}
TESTS_REQUIRE = [
    'pybel',
    'flask',
    'flask-admin',
]
ENTRY_POINTS = {
    'console_scripts': [
        'bio2bel = bio2bel.cli:main',
    ]
}

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """Build an absolute path from *parts* and return the contents of the resulting file. Assume UTF-8 encoding."""
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    meta_match = re.search(
        r'^__{meta}__ = ["\']([^"\']*)["\']'.format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string'.format(meta=meta))


def get_long_description():
    """Get the long_description from the README.rst file. Assume UTF-8 encoding."""
    with codecs.open(os.path.join(HERE, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


if __name__ == '__main__':
    setuptools.setup(
        name=find_meta('title'),
        version=find_meta('version'),
        description=find_meta('description'),
        long_description=get_long_description(),
        url=find_meta('url'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        maintainer='Charles Tapley Hoyt',
        maintainer_email=find_meta('email'),
        license=find_meta('license'),
        classifiers=CLASSIFIERS,
        keywords=KEYWORDS,
        packages=PACKAGES,
        package_dir={'': 'src'},
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRE,
        tests_require=TESTS_REQUIRE,
        entry_points=ENTRY_POINTS,
    )



# -*- coding: utf-8 -*-

"""Managers for Bio2BEL."""

from .abstract_manager import AbstractManager
from .connection_manager import ConnectionManager

__all__ = [
    'ConnectionManager',
    'AbstractManager',
]



# -*- coding: utf-8 -*-

"""Constants for Bio2BEL."""

from configparser import ConfigParser
import logging
import os

log = logging.getLogger(__name__)

VERSION = '0.1.1'

BIO2BEL_DIR = os.environ.get('BIO2BEL_DIRECTORY', os.path.join(os.path.expanduser('~'), '.pybel', 'bio2bel'))
os.makedirs(BIO2BEL_DIR, exist_ok=True)

DEFAULT_CONFIG_PATH = os.path.join(BIO2BEL_DIR, 'config.ini')

UNCONFIGURED_CACHE_NAME = 'bio2bel.db'
UNCONFIGURED_CACHE_PATH = os.path.join(BIO2BEL_DIR, UNCONFIGURED_CACHE_NAME)
UNCONFIGURED_CACHE_CONNECTION = 'sqlite:///' + UNCONFIGURED_CACHE_PATH


def get_global_connection():
    """Return the global connection string.

    :rtype: str
    """
    # 6. Check if there is a global connection
    global_environ_connection = os.environ.get('BIO2BEL_CONNECTION')
    if global_environ_connection is not None:
        log.info('loading global bio2bel connection from environ: %s', global_environ_connection)
        return global_environ_connection

    # 7. Use the global configuration file's global default cache connection string
    if not os.path.exists(DEFAULT_CONFIG_PATH):
        log.info('creating config file: %s', DEFAULT_CONFIG_PATH)
        config_writer = ConfigParser()
        with open(DEFAULT_CONFIG_PATH, 'w') as file:
            config_writer.set(config_writer.default_section, 'connection', UNCONFIGURED_CACHE_CONNECTION)
            config_writer.write(file)

    log.info('fetching global bio2bel config from %s', DEFAULT_CONFIG_PATH)
    config = ConfigParser()
    config.read(DEFAULT_CONFIG_PATH)

    if not config.has_option(config.default_section, 'connection'):
        log.info('creating default connection string %s', UNCONFIGURED_CACHE_CONNECTION)
        return UNCONFIGURED_CACHE_CONNECTION

    default_connection = config.get(config.default_section, 'connection')
    log.info('load default connection string from %s', default_connection)

    return default_connection


DEFAULT_CACHE_CONNECTION = get_global_connection()



# -*- coding: utf-8 -*-

"""Provide abstractions over BEL namespace generation procedures."""

from abc import ABC, abstractmethod
import logging
import sys
import time

import click
from tqdm import tqdm

from .cli_manager import CliMixin
from .connection_manager import ConnectionManager

log = logging.getLogger(__name__)

__all__ = [
    'Bio2BELMissingNamespaceModelError',
    'BELNamespaceManagerMixin',
]


class Bio2BELMissingNamespaceModelError(TypeError):
    """Raised when the namespace_model class variable is not defined."""


class BELNamespaceManagerMixin(ABC, ConnectionManager, CliMixin):
    """A mixin for generating a BEL namespace file and uploading it to the PyBEL database.

    First, you'll have to make sure that :mod:`pybel` is installed. This can be done with pip like:

    .. code-block:: bash

        $ pip install pybel

    To use this mixin, you need to properly implement the AbstractManager, and add additional class variables and
    functions.

    ``namespace_model``: The SQLAlchemy class that represents the entity to serialize into the namespace

    .. code-block:: python

        >>> from bio2bel import AbstractManager
        >>> from bio2bel.namespace_manager import NamespaceManagerMixin
        >>> from .models import HumanGene
        >>>
        >>> class MyManager(AbstractManager, NamespaceManagerMixin):
        ...     module_name = 'hgnc'
        ...     ...
        ...     namespace_model = HumanGene

    Several fields from Identifiers.org should be populated, including:

    1. ``identifiers_recommended``
    2. ``identifiers_pattern``
    3. ``identifiers_miriam``
    4. ``identifiers_namespace``
    5. ``identifiers_url``

    .. code-block:: python

        >>> from bio2bel import AbstractManager
        >>> from bio2bel.namespace_manager import NamespaceManagerMixin
        >>> from .models import HumanGene
        >>>
        >>> class MyManager(AbstractManager, NamespaceManagerMixin):
        ...     module_name = 'hgnc'
        ...     ...
        ...     namespace_model = HumanGene
        ...     identifiers_recommended = 'HGNC'
        ...     identifiers_pattern = '...'
        ...     identifiers_miriam = 'MIR:00000080'
        ...     identifiers_namespace = 'hgnc'
        ...     identifiers_url = 'http://identifiers.org/hgnc/'

    Two methods need to be implemented. First, the static method ``_get_identifier`` should take in the namespace model
    and give back the database identifier. for us, this is easy, since the HumanGene class has an attribute called
    ``hgnc_id``.

    Perhaps in the future, we will enfoce the convention that the namespace model should have a field
    called <module name>_id, but having this method gives lots of flexibility.

    This is also a good place to add more specific type annotations (not yet tested with MyPy).

    .. code-block:: python

        >>> from bio2bel import AbstractManager
        >>> from bio2bel.namespace_manager import NamespaceManagerMixin
        >>> from .models import HumanGene
        >>>
        >>> class MyManager(AbstractManager, NamespaceManagerMixin):
        ...     module_name = 'hgnc'
        ...     ...
        ...     namespace_model = HumanGene
        ...     identifiers_recommended = 'HGNC'
        ...     identifiers_pattern = '...'
        ...     identifiers_miriam = 'MIR:00000080'
        ...     identifiers_namespace = 'hgnc'
        ...     identifiers_url = 'http://identifiers.org/hgnc/'
        ...
        ...     @staticmethod
        ...     def _get_identifier(model: HumanGene) -> str:
        ...         return model.hgnc_id

    Last, we must implement the method ``_create_namespace_entry_from_model``, which encodes the logic of building a
    :class:`pybel.manager.models.NamespaceEntry` from the Bio2BEL repository's namespace model.

    For a repository like ChEBI, this is very simple, but for HGNC there is reason to add additional logic
    to get the proper encodings.

    .. code-block:: python

        >>> from bio2bel import AbstractManager
        >>> from bio2bel.namespace_manager import NamespaceManagerMixin
        >>> from pybel.manager.models import Namespace, NamespaceEntry
        >>> from .models import HumanGene
        >>>
        >>> class MyManager(AbstractManager, NamespaceManagerMixin):
        ...     module_name = 'hgnc'
        ...     ...
        ...     namespace_model = HumanGene
        ...     identifiers_recommended = 'HGNC'
        ...     identifiers_pattern = '...'
        ...     identifiers_miriam = 'MIR:00000080'
        ...     identifiers_namespace = 'hgnc'
        ...     identifiers_url = 'http://identifiers.org/hgnc/'
        ...
        ...     @staticmethod
        ...     def _get_identifier(model: HumanGene) -> str:
        ...         return model.hgnc_id
        ...
        ...     def _create_namespace_entry_from_model(self, model: HumanGene, namespace: Namespace) -> NamespaceEntry:
        ...         return NamespaceEntry(
        ...             encoding=encodings.get(model.locus_type, 'GRP'),
        ...             identifier=model.hgnc_id,
        ...             name=model.hgnc_symbol,
        ...             namespace=namespace,
        ...         )
    """

    namespace_model = ...

    identifiers_recommended = None
    identifiers_pattern = None
    identifiers_miriam = None
    identifiers_namespace = None
    identifiers_url = None

    def __init__(self, *args, **kwargs):  # noqa: D107
        if self.namespace_model is ...:
            raise Bio2BELMissingNamespaceModelError

        super().__init__(*args, **kwargs)

    @abstractmethod
    def _create_namespace_entry_from_model(self, model, namespace):
        """Create a PyBEL NamespaceEntry model from a Bio2BEL model.

        :param model: The model to convert
        :type namespace: pybel.manager.models.Namespace
        :rtype: Optional[pybel.manager.models.NamespaceEntry]
        """

    def _get_encoding(self, model):
        """Get the encoding for the model."""

    def _get_name(self, model):
        """Get the name for the namespace model."""

    @staticmethod
    @abstractmethod
    def _get_identifier(model):
        """Extract the identifier from an instance of namespace_model..

        :param model: The model to convert
        :rtype: str
        """

    def _iterate_namespace_models(self):
        """Return an iterator over the models to be converted to the namespace."""
        return tqdm(
            self._get_query(self.namespace_model),
            total=self._count_model(self.namespace_model),
            desc='Mapping {} to BEL namespace'.format(self._get_namespace_name())
        )

    @classmethod
    def _get_namespace_name(cls):
        """Get the nicely formatted name of this namespace.

        :rtype: str
        """
        return cls.identifiers_recommended or cls.module_name

    @classmethod
    def _get_namespace_keyword(cls):
        """Get the keyword to use as the reference BEL namespace.

        :rtype: str
        """
        return cls.identifiers_namespace or cls.module_name.upper()

    @classmethod
    def _get_namespace_url(cls):
        """Get the URL to use as the reference BEL namespace.

        :rtype: str
        """
        return cls.identifiers_url or '_{}'.format(cls.module_name.upper())

    def _get_default_namespace(self):
        """Get the reference BEL namespace if it exists.

        :rtype: Optional[pybel.manager.models.Namespace]
        """
        from pybel.manager.models import Namespace

        return self._get_query(Namespace).filter(Namespace.url == self._get_namespace_url()).one_or_none()

    def _get_namespace_entries(self, namespace):
        return [
            namespace_entry
            for namespace_entry in (
                self._create_namespace_entry_from_model(model, namespace)
                for model in self._iterate_namespace_models()
            )
            if namespace_entry is not None
        ]

    def _make_namespace(self):
        """Make a namespace.

        :rtype: pybel.manager.models.Namespace
        """
        from pybel.manager.models import Namespace

        namespace = Namespace(
            name=self._get_namespace_name(),
            keyword=self._get_namespace_keyword(),
            url=self._get_namespace_url(),
            version=str(time.asctime()),
        )
        self.session.add(namespace)

        entries = self._get_namespace_entries(namespace)
        self.session.add_all(entries)

        t = time.time()
        log.info('committing models')
        self.session.commit()
        log.info('committed models in %.2f seconds', time.time() - t)

        return namespace

    @staticmethod
    def _get_old_entry_identifiers(namespace):
        """Convert a PyBEL generalized namespace entries to a set.

        Default to using the identifier, but can be overridden to use the name instead.

        :param pybel.manager.models.Namespace namespace:
        :rtype: set[pybel.manager.model.NamespaceEntry]

        >>> {term.identifier for term in namespace.entries}
        """
        return {term.identifier for term in namespace.entries}

    def _update_namespace(self, namespace):
        """Update an already-created namespace.

        Note: Only call this if namespace won't be none!

        :type namespace: pybel.manager.models.Namespace
        """
        old_entry_identifiers = self._get_old_entry_identifiers(namespace)
        new_count = 0
        skip_count = 0

        for model in self._iterate_namespace_models():
            if self._get_identifier(model) in old_entry_identifiers:
                continue

            entry = self._create_namespace_entry_from_model(model, namespace=namespace)
            if entry is None or entry.name is None:
                skip_count += 1
                continue

            new_count += 1
            self.session.add(entry)

        t = time.time()
        log.info('got %d new entries. skipped %d entries missing names. committing models', new_count, skip_count)
        self.session.commit()
        log.info('committed models in %.2f seconds', time.time() - t)

    def upload_bel_namespace(self, update=False):
        """Upload the namespace to the PyBEL database.

        :param bool update: Should the namespace be updated first?
        :rtype: pybel.manager.models.Namespace
        """
        if not self.is_populated():
            self.populate()

        namespace = self._get_default_namespace()

        if namespace is None:
            log.info('making namespace for %s', self._get_namespace_name())
            return self._make_namespace()

        if update:
            self._update_namespace(namespace)

        return namespace

    def drop_bel_namespace(self):
        """Remove the default namespace if it exists.

        :rtype: Optional[pybel.manager.models.Namespace]
        """
        ns = self._get_default_namespace()

        if ns is not None:
            for entry in tqdm(ns.entries, desc='deleting entries in {}'.format(self._get_namespace_name())):
                self.session.delete(entry)
            self.session.delete(ns)

            log.info('committing deletions')
            self.session.commit()
            return ns

    def write_bel_namespace(self, file):
        """Write as a BEL namespace file.

        :param file:
        """
        from pybel.constants import NAMESPACE_DOMAIN_OTHER
        from pybel.resources import write_namespace

        values = {self._get_identifier(model) for model in self._iterate_namespace_models()}

        write_namespace(
            namespace_name=self._get_namespace_name(),
            namespace_keyword=self._get_namespace_keyword(),
            namespace_domain=NAMESPACE_DOMAIN_OTHER,
            namespace_query_url=self.identifiers_url,
            author_name='',
            citation_name='',
            values=values,
            functions='A',
            file=file,
        )

    @staticmethod
    def _cli_add_to_bel_namespace(main):
        """Add the export BEL namespace command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_to_bel_namespace(main)

    @staticmethod
    def _cli_add_clear_bel_namespace(main):
        """Add the clear BEL namespace command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_clear_bel_namespace(main)

    @staticmethod
    def _cli_add_write_bel_namespace(main):
        """Add the write BEL namespace command.

        :type main: click.Group
        :rtype: click.Group
        """
        return add_cli_write_bel_namespace(main)

    @classmethod
    def get_cli(cls):
        """Get a :mod:`click` main function with added BEL namespace commands.

        :rtype: click.Group
        """
        main = super().get_cli()

        @main.group()
        def belns():
            """Manage BEL namespace."""

        cls._cli_add_to_bel_namespace(belns)
        cls._cli_add_clear_bel_namespace(belns)
        cls._cli_add_write_bel_namespace(belns)

        return main


def add_cli_to_bel_namespace(main):
    """Add a ``upload_bel_namespace`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-u', '--update', is_flag=True)
    @click.pass_obj
    def upload(manager, update):
        """Upload names/identifiers to terminology store."""
        namespace = manager.upload_bel_namespace(update=update)
        click.echo('uploaded [{}] {}'.format(namespace.id, namespace.keyword))

    return main


def add_cli_clear_bel_namespace(main):
    """Add a ``clear_bel_namespace`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.pass_obj
    def drop(manager):
        """Clear names/identifiers to terminology store."""
        namespace = manager.drop_bel_namespace()

        if namespace:
            click.echo('namespace {} was cleared'.format(namespace))

    return main


def add_cli_write_bel_namespace(main):
    """Add a ``write_bel_namespace`` command to main :mod:`click` function.

    :param click.Group main: A click-decorated main function
    :rtype: click.Group
    """
    @main.command()
    @click.option('-f', '--file', type=click.File('w'), default=sys.stdout)
    @click.pass_obj
    def write(manager, file):
        """Write a BEL namespace names/identifiers to terminology store."""
        manager.write_bel_namespace(file)

    return main



# -*- coding: utf-8 -*-

"""Aggregate CLI for all Bio2BEL projects."""

import importlib
import logging
import os
import sys

import click
from pkg_resources import iter_entry_points, VersionConflict

from .constants import DEFAULT_CACHE_CONNECTION
from .models import Action
from .utils import get_version

log = logging.getLogger(__name__)

modules = {}
cli_modules = {}
main_commands = {}

for entry_point in iter_entry_points(group='bio2bel', name=None):
    entry = entry_point.name

    try:
        modules[entry] = entry_point.load()
    except VersionConflict:
        log.exception('Version conflict in %s', entry)
        continue
    except ImportError:
        log.exception('Issue with importing module %s', entry)
        continue

for entry, module in modules.items():
    try:
        cli_modules[entry] = modules[entry].cli
    except AttributeError:
        try:
            cli_modules[entry] = importlib.import_module('bio2bel_{}.cli'.format(entry))
        except ImportError:
            log.warning('no submodule bio2bel_%s.cli', entry)
            continue

    try:
        main_commands[entry] = cli_modules[entry].main
    except NameError:
        log.warning('no command group bio2bel_%s.cli:main', entry)
        continue

main = click.Group(commands=main_commands)
main.help = "Bio2BEL Command Line Utilities on {}\nBio2BEL v{}".format(sys.executable, get_version())


def _iterate_managers(connection, skip):
    _modules = sorted(
        (name, module)
        for name, module in modules.items()
        if name not in skip
    )
    return len(_modules), (
        (idx, name, module.Manager(connection=connection))
        for idx, (name, module) in enumerate(_modules, start=1)
    )


@main.command()
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
@click.option('--reset', is_flag=True, help='Nuke database first')
@click.option('--force', is_flag=True, help='Force overwrite if already populated')
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
def populate(connection, reset, force, skip):
    """Populate all."""
    lm, manager_list = _iterate_managers(connection, skip)

    for idx, name, manager in manager_list:
        click.echo(
            click.style('[{}/{}] '.format(idx, lm, name), fg='blue', bold=True) +
            click.style('populating {}'.format(name), fg='cyan', bold=True))

        if reset:
            click.echo('deleting the previous instance of the database')
            manager.drop_all()
            click.echo('creating new models')
            manager.create_all()

        elif manager.is_populated() and not force:
            click.echo('👍 {} is already populated. use --force to overwrite'.format(name), color='red')
            continue

        try:
            manager.populate()
        except Exception:
            log.exception('%s population failed', name)
            click.echo(click.style('👎 {} population failed'.format(name), fg='red', bold=True))


@main.command(help='Drop all')
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
def drop(connection, skip):
    """Drop all."""
    lm, manager_list = _iterate_managers(connection, skip)
    for idx, name, manager in manager_list:
        click.echo(click.style('dropping {}'.format(name), fg='cyan', bold=True))
        manager.drop_all()


@main.command()
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
def summarize(connection, skip):
    """Summarize all."""
    lm, manager_list = _iterate_managers(connection, skip)
    for idx, name, manager in manager_list:
        click.echo(click.style(name, fg='cyan', bold=True))
        if not manager.is_populated():
            click.echo('👎 unpopulated')
        elif not hasattr(manager, 'summarize'):
            click.echo('👎 summarize function not implemented')
        else:
            for field_name, count in sorted(manager.summarize().items()):
                click.echo(
                    click.style('=> ', fg='white', bold=True) +
                    '{}: {}'.format(field_name.replace('_', ' ').capitalize(), count)
                )


@main.command()
@click.option('-d', '--directory', type=click.Path(), default=os.getcwd(), help='output directory')
@click.option('--force', is_flag=True, help='Force overwrite if already exported')
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
def to_bel(directory, force, connection, skip):
    """Write all as BEL."""
    os.makedirs(directory, exist_ok=True)
    lm, manager_list = _iterate_managers(connection, skip)
    import pybel
    for idx, name, manager in manager_list:
        click.echo(click.style(name, fg='cyan', bold=True))
        path = os.path.join(directory, '{}.bel.gpickle'.format(name))
        if os.path.exists(path) and not force:
            click.echo('👍 already exported')
            continue

        if not manager.is_populated():
            click.echo('👎 unpopulated')
        elif not hasattr(manager, 'to_bel'):
            click.echo('👎 to_bel function not implemented')
        else:
            graph = manager.to_bel()
            pybel.to_pickle(graph, path)


@main.command()
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
def web(connection):
    """Run a combine web interface."""
    from bio2bel.web.application import create_application
    app = create_application(connection=connection)
    app.run(host='0.0.0.0', port=5000)


@main.command()
def actions():
    """List all actions."""
    for action in Action.ls():
        click.echo('{} {} {}'.format(action.created, action.action, action.resource))


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    main()
